package util;

import java.util.Vector;


import compilateur.CspVarManager;
import compilateur.Tool;
import compilateur.cspLangage.Type;

/**
 * VarManager gre un ensemble de variables et permet d'assurer l'unicit des
 * variables dans le programme.
 * 
 * @author Benjamin LE BOZEC
 */
abstract public class VarManager<VarType extends Var, VarTypeLinked extends Var> {

	/*
	 * DATA
	 */
	private Vector<VarType> varList = new Vector<VarType> ();
	private VarManager<VarTypeLinked, VarType> linkedManager = null;
	private VarManager<VarType, VarTypeLinked> parent = null;
	
	/*
	 * METHODS
	 */
	public VarManager () {
		
	}
	
	
	public VarManager (VarManager<VarType, VarTypeLinked> parent) {
		setParentManager(parent);
	}
	
	abstract protected VarType provideVar (String name, Type type, VarManager<VarType, VarTypeLinked> parent);
	
	/**
	 * Ajoute une variable sans faire intervenir un ventuel "linkedManager".
	 * La seule contrainte vrifie est l'unicit de la variable (si la contrainte
	 * n'est pas respecte, alors l'exception  VarAlreadyExistsException est leve).
	 * 
	 * @param name nom de la variable
	 * @param type type de la variable 
	 * @return Un objet "Var" englobant les ppts de la variable
	 * 
	 * @throws VarAlreadyExistsException leve si la variable existe dja
	 */
	public VarType addVariable (String name, Type type) throws VarAlreadyExistsException {
		if (isVarInside (name)) 
			throw new VarAlreadyExistsException ("La variable " + name + " est dja dclare");
		VarType v = provideVar(name, type, this);
		varList.add (v);
		return v;
	}
	
	/**
	 * Ajoute une variable dans la liste (cf addVariable) puis demande 
	 * l'attribution d'une variable dans le "linkedManager" afin de l'y associer.
	 * 
	 * @param name Nom de la  variable
	 * @param type Type de la variable
	 * @return L'objet Var cr
	 * @throws VarAlreadyExistsException Leve si la var existe dja dans la liste
	 */
	public VarType addLinkedVariable (String name, Type type, Type typeLinked) throws VarAlreadyExistsException {
		VarManager<VarTypeLinked, VarType> linkedManager = getLinkedManager();
		if (linkedManager == null) throw new NullPointerException ("Aucun linkedManager dclar !");
		if (typeLinked==null) throw new NullPointerException ("Attention : il est fortement conseill de dfinir un type pour CHAQUE variable");
		
		VarType v  = addVariable (name, type);
		v.linkTo(linkedManager.obtainVar(name, typeLinked));
		return v;
	}
	
	/**
	 * @param name Nom de la variable 
	 * @return VRAI si la variable est dclare dans la liste, FAUX sinon 
	 */
	public boolean isVarInside (String name) {
		return searchVar (name)!=null;
	}
	
	public boolean isVarInsideThis (String name) {
		return searchVarInThis(name)!=null;
	}
	/**
	 * Cette mthode permet l'attribution d'une variable en assurant l'unicit 
	 * de la variable ainsi attribue. Si une variable "name" existe dja, un
	 * autre nom lui sera attribu (tel que "name$0", etc...).
	 * 
	 * @param name Nom orignial de la variable (servant de base)
	 * @param type Type de la variable
	 * @return L'objet Var ainsi cr
	 * @throws VarAlreadyExistsException Leve si la var existe dja dans la liste
	 */
	public VarType obtainVar (String name, Type type) throws VarAlreadyExistsException {
		String search = name;
		Var found = null;
		while (searchVar(search)!=null)
			search = Tool.createNewName (search);
		/*do {
			found = null;
			for (Var var : varList) {
				if (var.getName().equals(search)) {
					found = var;
					search = Tool.createNewName(search);
				}
			}
		} while (found != null);*/
		return addVariable(search, type);
	}
	
	public void setLinkedManager (VarManager<VarTypeLinked, VarType> manager) {
		this.linkedManager = manager;
	}
	
	public void setParentManager (VarManager<VarType, VarTypeLinked> manager) {
		this.parent = manager;
		//setLinkedManager(manager.linkedManager);
	}
	
	public VarManager getParentManager () { return parent; }
	
	public String toString () {
		String descr = null;
		for (Var v : varList) {
			if (descr != null) descr += ", " + v.getName();
			else descr = v.getName();
		}
		return "[" + descr + "]";
	}

	protected VarType searchVarInThis (String name) {
		for (VarType var : varList)
			if (var.getName().equals(name)) return var;
		return null;
	}
	
	public VarType searchVar (String name) {
		VarType v = searchVarInThis(name);
		if (v!=null) return v;
		if (parent != null) return parent.searchVar(name);
		return null;
	}
	
	/**
	 * Retourne le "niveau" de la variable, c'est--dire le niveau suprieur
	 * auquel elle a t dclare. Retourne -1 si la var n'a pas t dclare
	 * 
	 * @param var
	 * @return le niveau, cad un nombre >= 0, ou -1 si la var n'a pas t dclare
	 */
	public int getVarLevel (Var var) {
		 Var v = searchVarInThis (var.getName());
		 if (v!=null) return 0;
		 if (parent!=null) return parent.getVarLevel0 (var, 1);
		 return -1;
	}
	
	private int getVarLevel0 (Var var, int level) {
		 Var v = searchVarInThis (var.getName());
		 if (v!=null) return level;
		 if (parent!=null) return parent.getVarLevel0 (var, level+1);
		 return -1;
	}

	public VarManager<VarTypeLinked,VarType> getThisLinkedManager() {
		return linkedManager;
	}

	public VarManager<VarTypeLinked, VarType> getLinkedManager () {
		if (linkedManager!=null) return linkedManager;
		if (parent!=null) return parent.getLinkedManager();
		return null;
	}

	public void setParentManagerOnly(VarManager<VarType, VarTypeLinked> cspVarList) {
		this.parent = cspVarList;
	}

	public void removeVar (String name) {
		VarType v = searchVar(name);
		removeVar (v);
	}
	
	public void removeVar (VarType var) {
		varList.remove(var);
		VarTypeLinked link = (VarTypeLinked) var.linkedVar;
		var.linkedVar = null;
		if (link!=null) {
			link.linkedVar = null;
			getLinkedManager().removeVar(link);
		}
	}
	
	protected Vector<VarType> getVarList () { return varList; }
}
